/******************************************************************************
 COMPARE.C

 Compares image to a lookup table.
 
 Subdivides the incoming image into 16 X 16 blocks, and calculates the mean 
 pixel intensity for each block.  The resultant array of means is compared to
 values from a lookup table to determine the number of blocks exceeding the
 threshold values in the lookup table.  A lookup table is selected based on
 the camera, filter, and exposure length, or a set of fixed values may be 
 specified.
 
 Accepts camera images with binning factors of 1, 2, or 4 in both axises.
 
 Writes the result of the comparision into a file for the decision module
 to read, and make a decision on which observation mode to execute next.

 Exits with a return value of -1 if there is any exception, or 0 if all 
 is well.
 
******************************************************************************/
#include <bios.h>
#include <dos.h>
#include <conio.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <io.h>
#include <sys\types.h>
#include <sys\stat.h>

#include "at200.h"
#include "idecls.h"
#include "camsys.h"

/* Function prototypes. */

void get_fname( void );
void get_lookup( void );
void show_help( void );
void scan_image( void );
void show_hdr( void );

/* Globals */

KI_DATA ki_data;	/* Sequence numbers, and name of last image file */

/* File header (fixed size 512 bytes) for file. */ 
 
union file_hdr_union 
{
      struct cam_structs_tag 
      { 	
         cam_misc misc;
         cam_sys  sys;
         cam_exp  exp;
      } structs;
      char     filler[512];
} file_hdr;
   
/* Actual image data. */

unsigned int img_data[32][512];		/* Raw input image (up to 32 rows)   */

/* 16 X 16 array of mean pixel value per block. */

unsigned int imavg[16][16];		

/* Lookup table. */

unsigned int lkup_val[16][16];

int pixel_max, pixel_min;
unsigned int nval, dispsw, cam_no;
unsigned int image_pbin, image_sbin, image_rows, image_cols, table;

char filespath[PATHSIZE], fname[80], lkupflnm[80];
int fd;

/* Get name of last image file that was written. */

void get_fname( void )
{
int	fdata, count;
char	data_fname[ 80 ];

  /* Open KI sequence data file. */

  strcpy ( data_fname, filespath );
  strcat ( data_fname, "KI_DATA.DAT" );
  fdata = open( data_fname, O_RDWR | O_BINARY , S_IREAD | S_IWRITE );
  if ( fdata < 0 )
  {
     	printf ( "Unable to open file %s.\n", data_fname );
     	exit( -1 );
  }

  /* Read the data record (only one). */

  count = read ( fdata, &ki_data, sizeof ( KI_DATA ) );
  if ( count < 1 )					/* Empty file ? */
  {
	printf( "ERROR: file %s is empty.\n", data_fname );
	exit( -1 );
  }
  
  /* pick up the image file name depending on camera # */ 

  if ( cam_no == 0 )
  	strcpy ( fname, ki_data.last_cam0_img_flnm );
  else
  	strcpy ( fname, ki_data.last_cam1_img_flnm );
  close ( fdata );
}

/* Read a data lookup table from a file that matches image. */
 
void get_lookup( void )
{
char	lkflname[13], databuff[80];
unsigned int expose_sec, expose_sel, line_no, ii, jj;
FILE 	*fp;

	/* Build file name for table lookup data.
	 * Name is CxFyXzzz.DAT where:
	 * x = camera #, 0 or 1
	 * y = filter #, 0 thru 4,
	 * zzz = exposure duration in seconds.
	 *       This is selected to 1, 10, or 60 second values.
	 */
	
	expose_sec = (unsigned int) (file_hdr.structs.exp.exposure / 1000L);
	if (expose_sec < 4 )
		expose_sel = 1;
	else if ( expose_sec < 30 )
		expose_sel = 10;
	else 
		expose_sel = 60;
		
	sprintf( lkflname, "C%dF%dX%3d.DAT", 
		file_hdr.structs.misc.camera,
		file_hdr.structs.misc.fw,
		expose_sel );
	for ( ii=0; ii<13; ii++ )		/* spaces to zeroes */
		if ( lkflname[ ii ] == ' ' ) 
			lkflname[ ii ] = '0'; 
	strcpy ( lkupflnm, filespath );		/* Build full pathname */
	strcat ( lkupflnm, lkflname );

	if ( dispsw )
		printf ( "Reading lookup table file: %s\n", lkupflnm );
	if ( ( fp=fopen( lkupflnm, "r" ) ) == NULL )
	{
		fprintf(stderr,"Can't open %s\n", lkupflnm );
		exit( -1 );
	}
	for ( line_no=0, ii=0; ii<16; ii++ )
	{
	   fgets( databuff, 80, fp );
	   line_no++;
	   jj = sscanf( &databuff[10], "%d %d %d %d %d %d %d %d",
		   &lkup_val[ii][0], &lkup_val[ii][1],
		   &lkup_val[ii][2],
		   &lkup_val[ii][3], &lkup_val[ii][4],
		   &lkup_val[ii][5],
		   &lkup_val[ii][6], &lkup_val[ii][7] );
	   if ( jj != 8 )
	   {
	   	printf("ERROR: Invalid value in threshold data file at line %d.\n",
	   		line_no );
	   	exit ( -1 );
	   }
	   fgets( databuff, 80, fp );
	   line_no++;
	   jj = sscanf( &databuff[10], "%d %d %d %d %d %d %d %d",
		   &lkup_val[ii][8], &lkup_val[ii][9], &lkup_val[ii][10],
		   &lkup_val[ii][11], &lkup_val[ii][12], &lkup_val[ii][13],
		   &lkup_val[ii][14], &lkup_val[ii][15] );
	   if ( jj != 8 )
	   {
	   	printf("ERROR: Invalid value in threshold data file at line %d.\n",
	   		line_no );
	   	exit ( -1 );
	   }
	}
	fclose ( fp );
}

void show_hdr ( void )
{
    printf ( "Image File: %s\n", fname );
    printf ( "MISC fields\n" );
    printf (  "preamp      = %4d       ", file_hdr.structs.misc.preamp ); 
    printf (  "ems_handle  = %4d\n",      file_hdr.structs.misc.ems_handle );
    printf (  "exp_scale   =%5u       ",  file_hdr.structs.misc.exp_scale );
    printf (  "exp_type    =    %c\n",    file_hdr.structs.misc.exposure_type );
    printf (  "tee         = %4d       ", file_hdr.structs.misc.temp.tee[0] );
    printf (  "td          = %4d\n",      file_hdr.structs.misc.temp.td[0] );
    printf (  "tcc         = %4d       ", file_hdr.structs.misc.temp.tcc[0] );
    printf (  "tcf         = %4d\n",      file_hdr.structs.misc.temp.tcf[0] );
    printf (  "tfw         = %4d       ", file_hdr.structs.misc.temp.tfw[0] );
    printf (  "tei         = %4d\n",      file_hdr.structs.misc.temp.tei[0] );
    printf (  "heater      = %4d       ", file_hdr.structs.misc.heater );
    printf (  "acpower     = %4d\n",      file_hdr.structs.misc.acpower );
    printf (  "pixel_min   =%5d       ",  file_hdr.structs.misc.pixel_min );
    printf (  "pixel_max   =%5d\n\n",     file_hdr.structs.misc.pixel_max );

    printf (  "CAM_EXP fields          " );
    printf (  "CAM_SYS fields\n" );
    printf (  "exposure    =%5ld      ",  file_hdr.structs.exp.exposure );
    printf (  "at200_id    = %4d\n",      file_hdr.structs.sys.at200_id );
    printf (  "gain        = %4d      ",  file_hdr.structs.exp.gain );
    printf (  "frame       = %4d\n",      file_hdr.structs.sys.frame );
    printf (  "corg        = %4d      ",  file_hdr.structs.exp.corg );
    printf (  "open_delay  = %4d\n",      file_hdr.structs.sys.open_delay );
    printf (  "rorg        = %4d      ",  file_hdr.structs.exp.rorg );
    printf (  "close_delay = %4d\n",      file_hdr.structs.sys.close_delay );
    printf (  "cols        = %4d      ",  file_hdr.structs.exp.cols );
    printf (  "preamp_delay= %4d\n",      file_hdr.structs.sys.preamp_delay );
    printf (  "rows        = %4d      ",  file_hdr.structs.exp.rows );
    printf (  "speed       = %4d\n",      file_hdr.structs.sys.speed );
    printf (  "sbin        = %4d      ",  file_hdr.structs.exp.sbin );
    printf (  "data_bits   = %4d\n",      file_hdr.structs.sys.data_bits );

    printf (  "pbin        = %4d\n",      file_hdr.structs.exp.pbin );
    printf (  "clr_cnt     = %4d\n",      file_hdr.structs.exp.clr_cnt );
    printf (  "row_shift   = %4d\n\n",    file_hdr.structs.exp.row_shift );
}

 
/* Read a camera image file, build 16 X 16 table of average pixel
 * intensity levels for comparasion to lookup table.
 */ 

void scan_image(void)
{
register unsigned row, col;
unsigned int	ii, jj, y, x, z, num, blkx, blky;
unsigned long	temp;

	/* Clear imavg table */
	
	for ( y=0; y<16; y++ )
		for ( x=0; x<16; x++ )
			imavg[y][x] = 0;

	/* Check if image to be processed is at an acceptable binning factor. */


	if ( !(   ( ( image_pbin == 1 ) && ( image_sbin == 1 ) )
	       || ( ( image_pbin == 2 ) && ( image_sbin == 2 ) )
	       || ( ( image_pbin == 4 ) && ( image_sbin == 4 ) ) ) )
	{
		printf("ERROR: binning must be 1X1, 2X2, or 4X4 for image.\n");
		exit ( -1 );
	}	      
	
	y = image_rows / image_pbin;	/* y = actual # of pixels in y axis */
	x = image_cols / image_sbin;	/* x = actual # of pixels in y axis */
	blky = y / 16;			/* blky = y pixels down   in block */
	blkx = x / 16;			/* blkx = x pixels across in block */
	num = blkx * blky;		/* number of pixels in block. */
	
	/* Process the input image data */

	for( row=0; row<y; row++ )       /* Each row of image data */
	{
		z = row % blky;
		read( fd, img_data[ z ], x*2 ); /* Read one row of data */
		if ( z == ( blky - 1 ) )	/* last row for this block ? */
		{
			/* Process this group of blocks */
			
			for( temp = 0L, col=0; col<x; col += blkx )
			{
			   for ( ii=0; ii<blky; ii++ ) /* blky pixels down */ 
			      for ( jj=0; jj<blkx; jj++ ) /* and blkx across */
			         temp += img_data[ii][col + jj];
			   temp += num / 2;
			   imavg[ (row / blky) ][ (col / blkx) ] = 
			   	(unsigned int) ( temp / num );
			   temp = 0L;
			}
		}
	}
}

void show_help(void)
{

  printf("USAGE: COMPARE [-f filename][-t nnn][-d][-n mmm][-l filename][-c x]\n"
         "  Where:\n"
         "    -f `filename' Name of the stored input image\n"
         "    -t  nnn       Use all 'nnn' values for table lookup data\n"
         "    -d            Display results\n"
         "    -n  mmm       Number of values that need to exceed threshold\n"
         "    -l 'filename' Name of table lookup file to use. (override)\n"
         "    -c  x         Use camera number 'x'. (default 0)\n" );
  exit( -1 );
}

void main(int argc,char *argv[])
{
unsigned int ii, jj, tval, x, y, line_no, count, decision;
FILE *fp;
char	databuff[82], *env;

	table = dispsw = cam_no = 0;
	nval = 128;		/* 50% of 16 X 16 blocks */
	fname[0] = '\0';

	/* Process command line arguments. */

	if ( ( env = getenv( "KI_FILES" ) ) != (char *) NULL )
	{
		strcpy( filespath, env );	/* Base path for KI files. */
		strcat( filespath, "\\" );
	}
	else
		filespath[0] = '\0';

	for( x=1 ; x < (unsigned) argc ; x++ ) 
	{
    	    if( argv[x][0] == '-' )
    		switch( argv[x][1] ) 
    		{

		/* Camera number. */

      		case 'c': case 'C':
      	  		cam_no = atoi( argv[ ++x ] );
      	  		if (( cam_no != 0 ) && ( cam_no != 1 ))
      	  			show_help();
      	  		break;
      	 
		/* Display output switch. */

      		case 'd': case 'D':
      	  		dispsw = 1;
      	  		break;
      	 
		/* Number of values that need to exceed threshold. */

      		case 'n': case 'N':
      	  		nval = atoi ( argv[ ++x ] );
      	  		break;
      	 
		/* Load lookup table with uniform value given by '-t nnn' arg. */

		case 't': case 'T':
      	  		tval = atoi ( argv[ ++x ] );
      	  		sprintf ( lkupflnm, "-t %s", argv[ x ] );
      	  		for ( ii=0; ii<16; ii++ )
      	     		    for ( jj=0; jj<16; jj++ )
				lkup_val[ii][jj] = tval;
	  		table = 1;    /* Show that lookup table is loaded. */
      	  		break;
      	  
			/* Use lookup table specified by '-l xxx" arg. */
			/* (overrides normal lookup table selection) */

        	case 'l': case 'L':
        		strcpy ( lkupflnm, argv[ ++x ] );
			if ( ( fp=fopen( lkupflnm, "r" ) ) == NULL )
			{
				fprintf(stderr,"Can't open %s.\n", argv[x] );
				exit( -1 );
			}
			for ( line_no=0, ii=0; ii<16; ii++ )
			{
		   	fgets( databuff, 80, fp );
		   	line_no++;
		   	jj = sscanf( &databuff[10], "%d %d %d %d %d %d %d %d",
			   &lkup_val[ii][0], &lkup_val[ii][1],
			   &lkup_val[ii][2],
			   &lkup_val[ii][3], &lkup_val[ii][4],
			   &lkup_val[ii][5],
			   &lkup_val[ii][6], &lkup_val[ii][7] );
		   	if ( jj != 8 )
		   	{
		   		printf("ERROR: Invalid value in threshold data file at line %d.\n",
		   			line_no );
		   		exit ( -1 );
		   	}
		   	fgets( databuff, 80, fp );
		   	line_no++;
		   	jj = sscanf( &databuff[10], "%d %d %d %d %d %d %d %d",
			   &lkup_val[ii][8], &lkup_val[ii][9], &lkup_val[ii][10],
			   &lkup_val[ii][11], &lkup_val[ii][12], &lkup_val[ii][13],
			   &lkup_val[ii][14], &lkup_val[ii][15] );
		   	if ( jj != 8 )
		   	{
		   		printf("ERROR: Invalid value in threshold data file at line %d.\n",
		   			line_no );
		   		exit ( -1 );
			}
			fclose ( fp );
			table = 1;
			break;

		case 'f': case 'F':		/* Image file to read. */
 			x++;
			strcpy ( fname, argv[x] );
         		break;

        	default:
			show_help();
        		break;
		}
	    }
	    else
		show_help();
	}
	
	/* Main line. */

	if ( !fname[0] )	/* Image file specified in command line? */
  		get_fname();	/* No - read it from KI_DATA file. */

        /* Open image data file. */

	fd = open( fname, O_RDONLY | O_BINARY , S_IREAD | S_IWRITE );
	if ( !fd )
	{
		printf( "ERROR: Unable to open file: %s\n", fname );
		exit( 1 );
	}
	
	count = read( fd, &file_hdr, 512 );	/* Read file header */
	if ( count == 512 )
	{
		if ( dispsw )
			show_hdr();
	}
	else
	{
		printf( "ERROR: Unable to read header for file: %s\n", fname );
		exit( -1 );
	}
	
	image_pbin = file_hdr.structs.exp.pbin;
	image_sbin = file_hdr.structs.exp.sbin;
	image_rows = file_hdr.structs.exp.rows;
	image_cols = file_hdr.structs.exp.cols;
	pixel_min  = file_hdr.structs.misc.pixel_min;
	pixel_max  = file_hdr.structs.misc.pixel_max;

  	if ( !table )		/* Lookup table loaded from command line? */
	  	get_lookup();   /* No - load it now. */
	
	
	scan_image();		/* Do it. */
	close(fd);

	/* Scan result table, compare to lookup table. */

	count = 0;
	if ( dispsw )
	{
		printf( "Results:\n\n" );

		printf( "------|" );
		for ( ii=0; ii<10; ii++ )
			printf("    %2d|", ii );
	}
	
	for ( y=0; y<16; y++ )
	{
		if ( dispsw )
			printf("\n%2d    |", y );
		for ( x=0; x<16; x++ )
		{
			if ( dispsw )
			{
				printf( " %5d ", imavg[y][x] );
				if ( !((x+1) % 10) )
					printf("\n      |" );
			}
			if ( imavg[y][x] > lkup_val[y][x] )
				count++;
		}
		if ( dispsw )
			printf("\n");
	}
	
	/* Set decision flag based on whether number of blocks exceeding
	 * lookup table threshold values exceeded 'nval'.
	 */

	if ( count > nval )
		decision = 1;
	else
		decision = 0;

	/* Write decision data to a file for majority decision logic 
	 * to analyze based on multiple images.
	 */  

	strcpy ( databuff, filespath );
	strcat ( databuff, "DECISION.DAT" );

	if ( ( fp=fopen( databuff, "a" ) ) == NULL ) /* Open for appending */
	{
		fprintf(stderr,"Can't open %s.\n", databuff );
		exit( -1 );
	}

	/* Add latest decision to file. */

	fprintf ( fp, "CAMERA %d, COUNT %3d, NVAL %3d, DECISION %d\n",
			file_hdr.structs.misc.camera,
			count, nval, decision, decision );
	fprintf ( fp, "          IMAGE  FILE: %s\n", fname );
	fprintf ( fp, "          LOOKUP FILE: %s\n", lkupflnm );
	fclose ( fp ); 

	/* Show it to user if display switch is on. */

	if ( dispsw )
	{
		printf ( "\nCAMERA %d, COUNT %3d, NVAL %3d, DECISION %d\n",
			file_hdr.structs.misc.camera,
			count, nval, decision, decision );
		printf ( "          IMAGE  FILE: %s\n", fname );
		printf ( "          LOOKUP FILE: %s\n", lkupflnm );
		
	}

	exit( 0 );	/* All OK */
}
